@@ -1,6 +1,7 @@
-#  Copyright (c) 2020 zfit
+#  Copyright (c) 2021 zfit
 
 import numpy as np
+
 import zfit
 
 # create space
@@ -8,7 +9,7 @@ obs = zfit.Space("x", limits=(-2, 3))
 
 # parameters
 mu = zfit.Parameter("mu", 1.2, -4, 6)
-sigma = zfit.Parameter("sigma", 1.3, 0.1, 10)
+sigma = zfit.Parameter("sigma", 1.3, 0.5, 10)
 
 # model building, pdf creation
 gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
@@ -20,9 +21,12 @@ data = zfit.Data.from_numpy(obs=obs, array=normal_np)
 # create NLL
 nll = zfit.loss.UnbinnedNLL(model=gauss, data=data)
 
-# create a minimizer
+# # create a minimizer
 minimizer = zfit.minimize.Minuit()
 result = minimizer.minimize(nll)
 
-# do the error calculations, here with minos
-param_errors, _ = result.errors()
+# do the error calculations with a hessian approximation
+param_errors = result.hesse()
+
+# or here with minos
+param_errors_asymetric, _ = result.errors()
