@@ -1,6 +1,9 @@
-"""Baseclass for a Model. Handle integration and sampling"""
+"""Baseclass for a Model.
 
-#  Copyright (c) 2020 zfit
+Handle integration and sampling
+"""
+
+#  Copyright (c) 2021 zfit
 
 import abc
 import builtins
@@ -9,29 +12,37 @@ import inspect
 import warnings
 from collections import OrderedDict
 from contextlib import suppress
-from typing import Dict, Union, Callable, List, Tuple, Optional
+from typing import Callable, Dict, List, Optional, Tuple, Union
 
 import tensorflow as tf
+from dotmap import DotMap
 from tensorflow_probability.python import mcmc as mc
 
-from . import integration as zintegrate, sample as zsample
-from .baseobject import BaseNumeric
-from .data import Data, Sampler, SampleData
-from .dependents import _extract_dependencies
-from .dimension import BaseDimensional
-from .interfaces import ZfitModel, ZfitParameter, ZfitData, ZfitSpace
-from .sample import UniformSampleAndWeights
-from .space import Space, convert_to_space, no_norm_range, supports
 from .. import z
 from ..core.integration import Integration
 from ..settings import ztypes
-from ..util import container as zcontainer, ztyping
+from ..util import container as zcontainer
+from ..util import ztyping
 from ..util.cache import GraphCachable
-from ..util.exception import (BasePDFSubclassingError, MultipleLimitsNotImplementedError, NormRangeNotImplementedError,
-                              ShapeIncompatibleError, SubclassingError, CannotConvertToNumpyError, WorkInProgressError,
-                              SpaceIncompatibleError, AnalyticIntegralNotImplementedError,
-                              SpecificFunctionNotImplementedError,
-                              AnalyticSamplingNotImplementedError, FunctionNotImplementedError)
+from ..util.exception import (AnalyticIntegralNotImplemented,
+                              AnalyticSamplingNotImplemented,
+                              BasePDFSubclassingError,
+                              CannotConvertToNumpyError,
+                              FunctionNotImplemented,
+                              MultipleLimitsNotImplemented,
+                              NormRangeNotImplemented, ShapeIncompatibleError,
+                              SpaceIncompatibleError,
+                              SpecificFunctionNotImplemented, SubclassingError,
+                              WorkInProgressError)
+from . import integration as zintegrate
+from . import sample as zsample
+from .baseobject import BaseNumeric
+from .data import Data, SampleData, Sampler
+from .dependents import _extract_dependencies
+from .dimension import BaseDimensional
+from .interfaces import ZfitData, ZfitModel, ZfitParameter, ZfitSpace
+from .sample import UniformSampleAndWeights
+from .space import Space, convert_to_space, no_norm_range, supports
 
 _BaseModel_USER_IMPL_METHODS_TO_CHECK = {}
 
@@ -42,7 +53,6 @@ def _BaseModel_register_check_support(has_support: bool):
     Args:
         has_support: If True, flags that it **requires** the `@supports` decorator. If False,
             flags that the `@supports` decorator is **not allowed**.
-
     """
     if not isinstance(has_support, bool):
         raise TypeError("Has to be boolean.")
@@ -68,9 +78,8 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     """Base class for any generic model.
 
     # TODO instructions on how to use
-
     """
-    _DEFAULTS_integration = zcontainer.DotDict()
+    _DEFAULTS_integration = DotMap()
     _DEFAULTS_integration.mc_sampler = lambda *args, **kwargs: mc.sample_halton_sequence(*args, randomized=False,
                                                                                          **kwargs)
     # _DEFAULTS_integration.mc_sampler = lambda dim, num_results, dtype: tf.random_uniform(maxval=1.,
@@ -97,7 +106,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         super().__init__(name=name, dtype=dtype, params=params, **kwargs)
         self._check_set_space(obs)
 
-        self._integration = zcontainer.DotDict()
+        self._integration = DotMap()
         self._integration.auto_numeric_integrator = self._DEFAULTS_integration.auto_numeric_integrator
         self.integration = Integration(mc_sampler=self._DEFAULTS_integration.mc_sampler,
                                        draws_per_dim=self._DEFAULTS_integration.draws_per_dim)
@@ -166,8 +175,17 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         self._space = obs.with_autofill_axes(overwrite=True)
 
     @contextlib.contextmanager
-    def _convert_sort_x(self, x: ztyping.XTypeInput, partial: bool = False) -> Data:
-        if isinstance(x, ZfitData):
+    def _convert_sort_x(self,
+                        x: ztyping.XTypeInput,
+                        partial: bool = False,
+                        allow_none: bool = False) -> Data:
+        if x is None:
+            if not allow_none:
+                raise ValueError(f"x {x} given to {self} must be non-empty (not None).")
+            else:
+                yield None
+
+        elif isinstance(x, ZfitData):
             if x.obs is not None:
                 with x.sort_by_obs(obs=self.obs, allow_superset=True):
                     yield x
@@ -218,7 +236,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
             self.integration.mc_sampler = mc_sampler
 
     # TODO: remove below? or add "analytic gradients"?
-    def gradients(self, x: ztyping.XType, norm_range: ztyping.LimitsType, params: ztyping.ParamsTypeOpt = None):
+    def gradient(self, x: ztyping.XType, norm_range: ztyping.LimitsType, params: ztyping.ParamsTypeOpt = None):
         raise NotImplementedError("Are the gradients needed?")
 
     def _check_input_norm_range(self, norm_range, none_is_error=False) -> Optional[ZfitSpace]:
@@ -232,7 +250,6 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
         Returns:
             Union[:py:class:`~zfit.Space`, False]:
-
         """
         if norm_range is None or (isinstance(norm_range, ZfitSpace) and not norm_range.limits_are_set):
             if none_is_error:
@@ -253,8 +270,8 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def convert_sort_space(self, obs: Union[ztyping.ObsTypeInput, ztyping.LimitsTypeInput] = None,
                            axes: ztyping.AxesTypeInput = None,
                            limits: ztyping.LimitsTypeInput = None) -> Union[ZfitSpace, None]:
-        """Convert the inputs (using eventually `obs`, `axes`) to :py:class:`~zfit.ZfitSpace` and sort them according to
-        own `obs`.
+        """Convert the inputs (using eventually `obs`, `axes`) to
+        :py:class:`~zfit.ZfitSpace` and sort them according to own `obs`.
 
         Args:
             obs:
@@ -262,7 +279,6 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
             limits:
 
         Returns:
-
         """
         if obs is None:  # for simple limits to convert them
             obs = self.obs
@@ -279,10 +295,14 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
     @_BaseModel_register_check_support(True)
     def _integrate(self, limits, norm_range):
-        raise SpecificFunctionNotImplementedError
+        raise SpecificFunctionNotImplemented
 
     @z.function(wraps='model')
-    def integrate(self, limits: ztyping.LimitsType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
+    def integrate(self,
+                  limits: ztyping.LimitsType,
+                  norm_range: ztyping.LimitsType = None,
+                  *,
+                  x: Optional[ztyping.DataInputType] = None) -> ztyping.XType:
         """Integrate the function over `limits` (normalized over `norm_range` if not False).
 
         Args:
@@ -295,7 +315,8 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         """
         norm_range = self._check_input_norm_range(norm_range)
         limits = self._check_input_limits(limits=limits)
-        integral = self._single_hook_integrate(limits=limits, norm_range=norm_range)
+        with self._convert_sort_x(x, allow_none=True) as x:
+            integral = self._single_hook_integrate(limits=limits, norm_range=norm_range, x=x)
         # TODO: allow integral values as arrays?
         # if isinstance(integral, tf.Tensor):
         #     if not integral.shape.as_list() == []:
@@ -307,7 +328,8 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         #                                      "instead of tensor)".format(integral.shape.as_list()))
         return integral
 
-    def _single_hook_integrate(self, limits, norm_range):
+    def _single_hook_integrate(self, limits, norm_range, x):
+        del x  # TODO HACK: how and what to pass through?
         return self._hook_integrate(limits=limits, norm_range=norm_range)
 
     def _hook_integrate(self, limits, norm_range):
@@ -316,7 +338,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _norm_integrate(self, limits, norm_range):
         try:
             integral = self._limits_integrate(limits=limits, norm_range=norm_range)
-        except NormRangeNotImplementedError:
+        except NormRangeNotImplemented:
             unnormalized_integral = self._limits_integrate(limits=limits, norm_range=False)
             normalization = self._limits_integrate(limits=norm_range, norm_range=False)
             integral = unnormalized_integral / normalization
@@ -325,7 +347,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _limits_integrate(self, limits, norm_range):
         try:
             integral = self._call_integrate(limits=limits, norm_range=norm_range)
-        except MultipleLimitsNotImplementedError:
+        except MultipleLimitsNotImplemented:
             integrals = []
             for sub_limits in limits:
                 integrals.append(self._call_integrate(limits=sub_limits, norm_range=norm_range))
@@ -334,9 +356,9 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
     def _call_integrate(self, limits, norm_range):
 
-        with suppress(FunctionNotImplementedError):
+        with suppress(FunctionNotImplemented):
             return self._integrate(limits=limits, norm_range=norm_range)
-        with suppress(AnalyticIntegralNotImplementedError):
+        with suppress(AnalyticIntegralNotImplemented):
             return self._hook_analytic_integrate(limits=limits, norm_range=norm_range)
         return self._fallback_integrate(limits=limits, norm_range=norm_range)
 
@@ -346,7 +368,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
         integral = None
         if max_axes and integral is None:  # TODO improve handling of available analytic integrals
-            with suppress(AnalyticIntegralNotImplementedError):
+            with suppress(AnalyticIntegralNotImplemented):
                 def part_int(x):
                     """Temporary partial integration function."""
                     return self._hook_partial_analytic_integrate(x, limits=limits, norm_range=norm_range)
@@ -383,7 +405,6 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
                 auto-handled.
             supports_norm_range: If `True`, `norm_range` argument to the function may not be `None`.
                 If `False`, `norm_range` will always be `None` and care is taken of the normalization automatically.
-
         """
         cls._analytic_integral.register(func=func, limits=limits, supports_norm_range=supports_norm_range,
                                         priority=priority, supports_multiple_limits=supports_multiple_limits)
@@ -403,7 +424,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
     @_BaseModel_register_check_support(True)
     def _analytic_integrate(self, limits, norm_range):
-        raise SpecificFunctionNotImplementedError
+        raise SpecificFunctionNotImplemented
 
     def analytic_integrate(self, limits: ztyping.LimitsType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
         """Analytical integration over function and raise Error if not possible.
@@ -433,19 +454,19 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _norm_analytic_integrate(self, limits, norm_range):
         try:
             integral = self._limits_analytic_integrate(limits=limits, norm_range=norm_range)
-        except NormRangeNotImplementedError:
+        except NormRangeNotImplemented:
 
             unnormalized_integral = self._limits_analytic_integrate(limits, norm_range=False)
             try:
                 normalization = self._limits_analytic_integrate(limits=norm_range, norm_range=False)
-            except (AnalyticIntegralNotImplementedError):
-                raise NormRangeNotImplementedError("Function does not support this (or even any)"
-                                                   "normalization range 'norm_range'."
-                                                   " This usually means,that no analytic integral "
-                                                   "is available for this function. Due to rule "
-                                                   "safety, an analytical normalization has to "
-                                                   "be available and no attempt of numerical "
-                                                   "normalization was made.")
+            except (AnalyticIntegralNotImplemented):
+                raise NormRangeNotImplemented("Function does not support this (or even any)"
+                                              "normalization range 'norm_range'."
+                                              " This usually means,that no analytic integral "
+                                              "is available for this function. Due to rule "
+                                              "safety, an analytical normalization has to "
+                                              "be available and no attempt of numerical "
+                                              "normalization was made.")
             else:
                 integral = unnormalized_integral / normalization
         return integral
@@ -453,7 +474,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _limits_analytic_integrate(self, limits, norm_range):
         try:
             integral = self._call_analytic_integrate(limits, norm_range=norm_range)
-        except MultipleLimitsNotImplementedError:
+        except MultipleLimitsNotImplemented:
             integrals = []
             for sub_limits in limits:
                 integrals.append(self._call_analytic_integrate(limits=sub_limits, norm_range=norm_range))
@@ -461,7 +482,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         return integral
 
     def _call_analytic_integrate(self, limits, norm_range):
-        with suppress(FunctionNotImplementedError, AnalyticIntegralNotImplementedError):
+        with suppress(FunctionNotImplemented, AnalyticIntegralNotImplemented):
             return self._analytic_integrate(limits=limits, norm_range=norm_range)
         return self._fallback_analytic_integrate(limits=limits, norm_range=norm_range)
 
@@ -469,12 +490,12 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         try:
             return self._analytic_integral.integrate(x=None, limits=limits, axes=limits.axes,
                                                      norm_range=norm_range, model=self, params=self.params)
-        except (SpecificFunctionNotImplementedError, AnalyticIntegralNotImplementedError):
-            raise AnalyticIntegralNotImplementedError
+        except (SpecificFunctionNotImplemented, AnalyticIntegralNotImplemented):
+            raise AnalyticIntegralNotImplemented
 
     @_BaseModel_register_check_support(True)
     def _numeric_integrate(self, limits, norm_range):
-        raise SpecificFunctionNotImplementedError
+        raise SpecificFunctionNotImplemented
 
     def numeric_integrate(self, limits: ztyping.LimitsType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
         """Numerical integration over the model.
@@ -485,7 +506,6 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
         Returns:
             The integral value
-
         """
         norm_range = self._check_input_norm_range(norm_range)
         limits = self._check_input_limits(limits=limits)
@@ -501,7 +521,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _norm_numeric_integrate(self, limits, norm_range):
         try:
             integral = self._limits_numeric_integrate(limits=limits, norm_range=norm_range)
-        except NormRangeNotImplementedError:
+        except NormRangeNotImplemented:
             assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
             unnormalized_integral = self._limits_numeric_integrate(limits=limits, norm_range=False)
             normalization = self._limits_numeric_integrate(limits=norm_range, norm_range=False)
@@ -511,7 +531,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _limits_numeric_integrate(self, limits, norm_range):
         try:
             integral = self._call_numeric_integrate(limits=limits, norm_range=norm_range)
-        except MultipleLimitsNotImplementedError:
+        except MultipleLimitsNotImplemented:
             integrals = []
             for sub_limits in limits:
                 integrals.append(self._call_numeric_integrate(limits=sub_limits, norm_range=norm_range))
@@ -520,7 +540,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         return integral
 
     def _call_numeric_integrate(self, limits, norm_range):
-        with suppress(FunctionNotImplementedError):
+        with suppress(FunctionNotImplemented):
             return self._numeric_integrate(limits=limits, norm_range=norm_range)
         return self._fallback_numeric_integrate(limits=limits, norm_range=norm_range)
 
@@ -529,7 +549,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
     @_BaseModel_register_check_support(True)
     def _partial_integrate(self, x, limits, norm_range):
-        raise SpecificFunctionNotImplementedError
+        raise SpecificFunctionNotImplemented
 
     @z.function(wraps='model')
     def partial_integrate(self, x: ztyping.XTypeInput, limits: ztyping.LimitsType,
@@ -562,7 +582,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _norm_partial_integrate(self, x, limits, norm_range):
         try:
             integral = self._limits_partial_integrate(x=x, limits=limits, norm_range=norm_range)
-        except NormRangeNotImplementedError:
+        except NormRangeNotImplemented:
             assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
             unnormalized_integral = self._limits_partial_integrate(x=x, limits=limits, norm_range=False)
             normalization = self._hook_integrate(limits=norm_range, norm_range=False)
@@ -572,7 +592,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _limits_partial_integrate(self, x, limits, norm_range):
         try:
             integral = self._call_partial_integrate(x=x, limits=limits, norm_range=norm_range)
-        except MultipleLimitsNotImplementedError:
+        except MultipleLimitsNotImplemented:
             integrals = []
             for sub_limit in limits:
                 integrals.append(self._call_partial_integrate(x=x, limits=sub_limit, norm_range=norm_range))
@@ -582,14 +602,14 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
     def _call_partial_integrate(self, x, limits, norm_range):
 
-        with suppress(FunctionNotImplementedError):
+        with suppress(FunctionNotImplemented):
             return self._partial_integrate(x=x, limits=limits, norm_range=norm_range)
-        with suppress(AnalyticIntegralNotImplementedError):
+        with suppress(AnalyticIntegralNotImplemented):
             return self._hook_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
         try:
             return self._fallback_partial_integrate(x=x, limits=limits, norm_range=norm_range)
-        except FunctionNotImplementedError:
-            raise AnalyticIntegralNotImplementedError
+        except FunctionNotImplemented:
+            raise AnalyticIntegralNotImplemented
 
     def _fallback_partial_integrate(self, x, limits: ZfitSpace, norm_range: ZfitSpace):
         max_axes = self._analytic_integral.get_max_axes(limits=limits, axes=limits.axes)
@@ -612,7 +632,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
 
     @_BaseModel_register_check_support(True)
     def _partial_analytic_integrate(self, x, limits, norm_range):
-        raise SpecificFunctionNotImplementedError
+        raise SpecificFunctionNotImplemented
 
     @z.function(wraps='model')
     def partial_analytic_integrate(self, x: ztyping.XTypeInput, limits: ztyping.LimitsType,
@@ -635,7 +655,6 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
             NormRangeNotImplementedError: if the *norm_range* argument is not supported. This
                 means that no analytical normalization is available, explicitly: the **analytical**
                 integral over the limits = norm_range is not available.
-
         """
         norm_range = self._check_input_norm_range(norm_range=norm_range)
         limits = self._check_input_limits(limits=limits)
@@ -651,16 +670,16 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _norm_partial_analytic_integrate(self, x, limits, norm_range):
         try:
             integral = self._limits_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
-        except NormRangeNotImplementedError:
+        except NormRangeNotImplemented:
             assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
             unnormalized_integral = self._limits_partial_analytic_integrate(x=x, limits=limits, norm_range=False)
             try:
                 normalization = self._limits_analytic_integrate(limits=norm_range, norm_range=False)
-            except AnalyticIntegralNotImplementedError:
-                raise NormRangeNotImplementedError("Function does not support this (or even any) normalization range"
-                                                   " 'norm_range'. This usually means,that no analytic integral "
-                                                   "is available for this function. An analytical normalization has to "
-                                                   "be available and no attempt of numerical normalization was made.")
+            except AnalyticIntegralNotImplemented:
+                raise NormRangeNotImplemented("Function does not support this (or even any) normalization range"
+                                              " 'norm_range'. This usually means,that no analytic integral "
+                                              "is available for this function. An analytical normalization has to "
+                                              "be available and no attempt of numerical normalization was made.")
             else:
                 integral = unnormalized_integral / normalization
         return integral
@@ -668,7 +687,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _limits_partial_analytic_integrate(self, x, limits, norm_range):
         try:
             integral = self._call_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
-        except MultipleLimitsNotImplementedError:
+        except MultipleLimitsNotImplemented:
             integrals = []
             for sub_limits in limits:
                 integrals.append(self._call_partial_analytic_integrate(x=x, limits=sub_limits, norm_range=norm_range))
@@ -677,7 +696,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         return integral
 
     def _call_partial_analytic_integrate(self, x, limits, norm_range):
-        with suppress(FunctionNotImplementedError, AnalyticIntegralNotImplementedError):
+        with suppress(FunctionNotImplemented, AnalyticIntegralNotImplemented):
             return self._partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
         return self._fallback_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
 
@@ -685,12 +704,12 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         try:
             return self._analytic_integral.integrate(x=x, limits=limits, axes=limits.axes,
                                                      norm_range=norm_range, model=self, params=self.params)
-        except (SpecificFunctionNotImplementedError, AnalyticIntegralNotImplementedError):
-            raise AnalyticIntegralNotImplementedError
+        except (SpecificFunctionNotImplemented, AnalyticIntegralNotImplemented):
+            raise AnalyticIntegralNotImplemented
 
     @_BaseModel_register_check_support(True)
     def _partial_numeric_integrate(self, x, limits, norm_range):
-        raise SpecificFunctionNotImplementedError
+        raise SpecificFunctionNotImplemented
 
     @z.function(wraps='model')
     def partial_numeric_integrate(self, x: ztyping.XType, limits: ztyping.LimitsType,
@@ -723,7 +742,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _norm_partial_numeric_integrate(self, x, limits, norm_range):
         try:
             integral = self._limits_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
-        except NormRangeNotImplementedError:
+        except NormRangeNotImplemented:
             assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
             unnormalized_integral = self._limits_partial_numeric_integrate(x=x, limits=limits, norm_range=False)
             integral = unnormalized_integral / self._hook_numeric_integrate(limits=norm_range, norm_range=norm_range)
@@ -732,7 +751,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     def _limits_partial_numeric_integrate(self, x, limits, norm_range):
         try:
             integral = self._call_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
-        except MultipleLimitsNotImplementedError:
+        except MultipleLimitsNotImplemented:
             integrals = []
             for sub_limits in limits:
                 integrals.append(self._call_partial_numeric_integrate(x=x, limits=sub_limits, norm_range=norm_range))
@@ -740,7 +759,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         return integral
 
     def _call_partial_numeric_integrate(self, x, limits, norm_range):
-        with suppress(SpecificFunctionNotImplementedError):
+        with suppress(SpecificFunctionNotImplemented):
             return self._partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
         return self._fallback_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
 
@@ -762,7 +781,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     @supports()
     def _inverse_analytic_integrate(self, x):
         if not self._inverse_analytic_integral:
-            raise AnalyticSamplingNotImplementedError
+            raise AnalyticSamplingNotImplemented
         else:
             icdf = self._inverse_analytic_integral[0]
             params = inspect.signature(icdf).parameters
@@ -796,7 +815,7 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
                 been created with when it resamples.
 
         Returns:
-            :py:class:~`zfit.core.data.Sampler`
+            :py:class:`~zfit.core.data.Sampler`
 
         Raises:
             NotExtendedPDFError: if 'extended' is chosen (implicitly by default or explicitly) as an
@@ -830,15 +849,17 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
     @z.function(wraps='model')
     def _create_sampler_tensor(self, limits, n):
 
-        sample = self._single_hook_sample(n=n, limits=limits)
+        sample = self._single_hook_sample(n=n, limits=limits, x=None)
         return sample
 
     @_BaseModel_register_check_support(True)
     def _sample(self, n, limits: ZfitSpace):
-        raise SpecificFunctionNotImplementedError
+        raise SpecificFunctionNotImplemented
 
     def sample(self, n: ztyping.nSamplingTypeIn = None,
-               limits: ztyping.LimitsType = None) -> SampleData:  # TODO: change poissonian top-level with multinomial
+               limits: ztyping.LimitsType = None,
+               x: Optional[
+                   ztyping.DataInputType] = None) -> SampleData:  # TODO: change poissonian top-level with multinomial
         """Sample `n` points within `limits` from the model.
 
         If `limits` is not specified, `space` is used (if the space contains limits).
@@ -872,35 +893,42 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         limits = self._check_input_limits(limits=limits, none_is_error=True)
 
         @z.function(wraps='model_sampling')
-        def run_tf(n, limits):
-            sample = self._single_hook_sample(n=n, limits=limits)
+        def run_tf(n, limits, x):
+            sample = self._single_hook_sample(n=n, limits=limits, x=x)
             return sample
 
-        sample_data = SampleData.from_sample(sample=run_tf(n=n, limits=limits), obs=limits)
+        with self._convert_sort_x(x, allow_none=True) as x:
+            if x is not None:
+                new_obs = limits * x.data_range
+            else:
+                new_obs = limits
+            sample_data = SampleData.from_sample(sample=run_tf(n=n, limits=limits, x=x),
+                                                 obs=new_obs)  # TODO: which limits?
 
         return sample_data
 
-    def _single_hook_sample(self, n, limits):
+    def _single_hook_sample(self, n, limits, x=None):
+        del x
         return self._hook_sample(n=n, limits=limits)
 
     def _hook_sample(self, limits, n):
         return self._norm_sample(n=n, limits=limits)
 
     def _norm_sample(self, n, limits):
-        """Dummy function"""
+        """Dummy function."""
         return self._limits_sample(n=n, limits=limits)
 
     def _limits_sample(self, n, limits):
         try:
             return self._call_sample(n=n, limits=limits)
-        except MultipleLimitsNotImplementedError as error:
+        except MultipleLimitsNotImplemented as error:
             try:
                 total_integral = self.analytic_integrate(limits, norm_range=False)
                 sub_integrals = tf.concat([self.analytic_integrate(limit, norm_range=False) for limit in limits],
                                           axis=0)
-            except AnalyticIntegralNotImplementedError:
-                raise MultipleLimitsNotImplementedError("Cannot autohandle multiple limits as the analytic"
-                                                        " integral is not available.") from error
+            except AnalyticIntegralNotImplemented:
+                raise MultipleLimitsNotImplemented("Cannot autohandle multiple limits as the analytic"
+                                                   " integral is not available.") from error
             fracs = sub_integrals / total_integral
             n_samples = tf.unstack(z.random.counts_multinomial(n, probs=fracs), axis=0)
 
@@ -915,17 +943,17 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
         return sample
 
     def _call_sample(self, n, limits):
-        with suppress(SpecificFunctionNotImplementedError):
+        with suppress(SpecificFunctionNotImplemented):
             return self._sample(n=n, limits=limits)
-        with suppress(SpecificFunctionNotImplementedError, AnalyticSamplingNotImplementedError):
+        with suppress(SpecificFunctionNotImplemented, AnalyticSamplingNotImplemented):
             return self._analytic_sample(n=n, limits=limits)
         return self._fallback_sample(n=n, limits=limits)
 
     def _analytic_sample(self, n, limits: ZfitSpace):  # TODO(Mayou36) implement multiple limits sampling
         if not self._inverse_analytic_integral:
-            raise AnalyticSamplingNotImplementedError  # TODO(Mayou36): create proper analytic sampling
+            raise AnalyticSamplingNotImplemented  # TODO(Mayou36): create proper analytic sampling
         if limits.n_limits > 1:
-            raise AnalyticSamplingNotImplementedError
+            raise AnalyticSamplingNotImplemented
         try:
             lower_bound, upper_bound = limits.rect_limits_np
         except CannotConvertToNumpyError as err:
@@ -943,9 +971,9 @@ class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
                                                                                 (upper_bound,)),
                                                                         axes=limits.axes),
                                                            norm_range=False)
-        except (SpecificFunctionNotImplementedError, AnalyticIntegralNotImplementedError):
-            raise AnalyticSamplingNotImplementedError(f"analytic sampling not possible because the analytic integral"
-                                                      f" is not"" implemented for the boundaries: {limits}")
+        except (SpecificFunctionNotImplemented, AnalyticIntegralNotImplemented):
+            raise AnalyticSamplingNotImplemented(f"analytic sampling not possible because the analytic integral"
+                                                 f" is not"" implemented for the boundaries: {limits}")
         prob_sample = z.random.uniform(shape=(n, limits.n_obs), minval=lower_prob_lim,
                                        maxval=upper_prob_lim)
         # with self._convert_sort_x(prob_sample) as x:
@@ -1054,7 +1082,7 @@ class SimpleModelSubclassMixin:
                 sigma = self.params['sigma']
                 x = z.unstack_x(x)
                 return z.exp(-z.square((x - mu) / sigma))
-        """
+    """
 
     def __init__(self, *args, **kwargs):
         try:
@@ -1076,4 +1104,4 @@ class SimpleModelSubclassMixin:
                                    "    _PARAMS = ['mu', 'sigma']")
         not_str = [param for param in params if not isinstance(param, str)]
         if not_str:
-            raise TypeError("The following parameters are not strings in `_PARAMS`: ".format(not_str))
+            raise TypeError(f"The following parameters are not strings in `_PARAMS`: ")
