@@ -1,14 +1,14 @@
-#  Copyright (c) 2020 zfit
+#  Copyright (c) 2021 zfit
 
 import abc
 from abc import ABCMeta, abstractmethod
-from typing import Union, List, Dict, Callable, Tuple, Optional, Set
+from typing import Callable, Dict, List, Optional, Set, Tuple, Union
 
 import numpy as np
 import tensorflow as tf
-from ..util.deprecation import deprecated
 
 from ..util import ztyping
+from ..util.deprecation import deprecated
 
 
 class ZfitObject(abc.ABC):
@@ -35,7 +35,10 @@ class ZfitDimensional(ZfitObject):
     @property
     @abstractmethod
     def n_obs(self) -> int:
-        """Return the number of observables, the dimensionality. Corresponds to the last dimension."""
+        """Return the number of observables, the dimensionality.
+
+        Corresponds to the last dimension.
+        """
         raise NotImplementedError
 
 
@@ -88,36 +91,36 @@ class ZfitOrderableDimensional(ZfitDimensional, metaclass=ABCMeta):
                   allow_subset: bool = True) -> "ZfitOrderableDimensional":
         """Create a new instance that has `axes`; sorted by or set or dropped.
 
-            The behavior is as follows:
+        The behavior is as follows:
 
-             * axes are already set:
-               * input axes are None: the axes will be dropped. If no observables are set, an error
-                 will be raised, as no coordinates will be assigned to this instance anymore.
-               * input axes are not None: the instance will be sorted by the incoming axes. If obs or other
-                 objects have an associated order (e.g. data, limits,...), they will be reordered as well.
-                 If a strict subset is given (and allow_subset is True), only a subset will be returned. This can
-                 be used to retrieve a subspace of limits, data etc.
-                 If a strict superset is given (and allow_superset is True), the axes will be sorted accordingly as
-                 if the axes not contained in the instances axes were not present in the input axes.
-             * axes are not set:
-               * if the input axes are None, the same object is returned.
-               * if the input axes are not None, they will be set as-is and now correspond to the already
-                 existing obs in the object.
+         * axes are already set:
+           * input axes are None: the axes will be dropped. If no observables are set, an error
+             will be raised, as no coordinates will be assigned to this instance anymore.
+           * input axes are not None: the instance will be sorted by the incoming axes. If obs or other
+             objects have an associated order (e.g. data, limits,...), they will be reordered as well.
+             If a strict subset is given (and allow_subset is True), only a subset will be returned. This can
+             be used to retrieve a subspace of limits, data etc.
+             If a strict superset is given (and allow_superset is True), the axes will be sorted accordingly as
+             if the axes not contained in the instances axes were not present in the input axes.
+         * axes are not set:
+           * if the input axes are None, the same object is returned.
+           * if the input axes are not None, they will be set as-is and now correspond to the already
+             existing obs in the object.
 
-            Args:
-                axes: Axes to sort/associate this instance with
-                allow_superset: if False and a strict superset of the own axeservables is given, an error
-                is raised.
-                allow_subset:if False and a strict subset of the own axeservables is given, an error
-                is raised.
+        Args:
+            axes: Axes to sort/associate this instance with
+            allow_superset: if False and a strict superset of the own axeservables is given, an error
+            is raised.
+            allow_subset:if False and a strict subset of the own axeservables is given, an error
+            is raised.
 
-            Returns:
-                A copy of the object with the new ordering/axes
+        Returns:
+            A copy of the object with the new ordering/axes
 
-            Raises:
-                CoordinatesUnderdefinedError: if obs is None and the instance does not have axes
-                AxesIncompatibleError: if `axes` is a superset and allow_superset is False or a subset and
-                    allow_allow_subset is False
+        Raises:
+            CoordinatesUnderdefinedError: if obs is None and the instance does not have axes
+            AxesIncompatibleError: if `axes` is a superset and allow_superset is False or a subset and
+                allow_allow_subset is False
         """
 
         raise NotImplementedError
@@ -299,7 +302,10 @@ class ZfitLimit(abc.ABC, metaclass=ABCMeta):
 
     @abstractmethod
     def rect_area(self) -> Union[float, np.ndarray, tf.Tensor]:
-        """Calculate the total rectangular area of all the limits and axes. Useful, for example, for MC integration."""
+        """Calculate the total rectangular area of all the limits and axes.
+
+        Useful, for example, for MC integration.
+        """
         raise NotImplementedError
 
     @abstractmethod
@@ -420,6 +426,7 @@ class ZfitLimit(abc.ABC, metaclass=ABCMeta):
     @abstractmethod
     def __eq__(self, other: object) -> bool:
         """Compares two Limits for equality without graph mode allowed.
+
         Raises:
              IllegalInGraphModeError: it the comparison happens with tensors in a graph context.
         """
@@ -445,7 +452,6 @@ class ZfitLimit(abc.ABC, metaclass=ABCMeta):
             Result of the comparison
         Raises:
              IllegalInGraphModeError: it the comparison happens with tensors in a graph context.
-
         """
         raise NotImplementedError
 
@@ -493,24 +499,23 @@ class ZfitSpace(ZfitLimit, ZfitOrderableDimensional, ZfitObject, metaclass=ABCMe
     @property
     @abstractmethod
     def lower(self) -> ztyping.LowerTypeReturn:
-        """Return the lower limits.
-
-        """
+        """Return the lower limits."""
         raise NotImplementedError
 
     # TODO: legacy?
     @property
     @abstractmethod
     def upper(self) -> ztyping.UpperTypeReturn:
-        """Return the upper limits.
-
-        """
+        """Return the upper limits."""
         raise NotImplementedError
 
     # TODO: legacy?
     @abstractmethod
     def area(self) -> float:
-        """Return the total area of all the limits and axes. Useful, for example, for MC integration."""
+        """Return the total area of all the limits and axes.
+
+        Useful, for example, for MC integration.
+        """
         raise NotImplementedError
 
     @abstractmethod
@@ -542,7 +547,6 @@ class ZfitSpace(ZfitLimit, ZfitOrderableDimensional, ZfitObject, metaclass=ABCMe
             name:
 
         Returns:
-
         """
         raise NotImplementedError
 
@@ -626,7 +630,7 @@ class ZfitParametrized(ZfitDependenciesMixin, ZfitObject):
 
     @property
     @abstractmethod
-    def params(self) -> ztyping.ParametersType:
+    def params(self) -> ztyping.ParameterType:
         raise NotImplementedError
 
 
@@ -681,7 +685,6 @@ class ZfitIndependentParameter(ZfitParameter, metaclass=ABCMeta):
     def randomize(self, minval, maxval, sampler):
         """Update the parameter with a randomised value between minval and maxval and return it.
 
-
         Args:
             minval: The lower bound of the sampler. If not given, `lower_limit` is used.
             maxval: The upper bound of the sampler. If not given, `upper_limit` is used.
@@ -717,7 +720,6 @@ class ZfitIndependentParameter(ZfitParameter, metaclass=ABCMeta):
 
         Returns:
             Boolean `tf.Tensor` that tells whether the value is at the limits.
-
         """
         raise NotImplementedError
 
@@ -739,18 +741,13 @@ class ZfitIndependentParameter(ZfitParameter, metaclass=ABCMeta):
 class ZfitLoss(ZfitObject, metaclass=ABCMeta):
 
     @abstractmethod
-    def gradients(self, params: ztyping.ParamTypeInput = None) -> List[tf.Tensor]:
+    def gradient(self, params: ztyping.ParamTypeInput = None) -> List[tf.Tensor]:
         raise NotImplementedError
 
     @abstractmethod
     def value(self) -> ztyping.NumericalTypeReturn:
         raise NotImplementedError
 
-    @property
-    @abstractmethod
-    def errordef(self) -> Union[float, int]:
-        raise NotImplementedError
-
     @property
     @abstractmethod
     def model(self) -> List["ZfitModel"]:
@@ -775,12 +772,15 @@ class ZfitLoss(ZfitObject, metaclass=ABCMeta):
     def errordef(self) -> float:
         raise NotImplementedError
 
-    @abstractmethod
-    def value_gradients(self, params):
+    def hessian(self, params):
         pass
 
     @abstractmethod
-    def value_gradients_hessian(self, params, hessian=None):
+    def value_gradient(self, params):
+        pass
+
+    @abstractmethod
+    def value_gradient_hessian(self, params, hessian=None):
         pass
 
 
@@ -822,7 +822,6 @@ class ZfitModel(ZfitNumericParametrized, ZfitDimensional):
             supports_norm_range:
 
         Returns:
-
         """
         raise NotImplementedError
 
